import { HttpException, HttpStatus, Injectable } from "@nestjs/common";
import { InjectRepository } from "@nestjs/typeorm";
import { Constants } from "src/utils/constants";
import { ResponseUtils } from "src/utils/response.utils";
import { DoSpacesService } from "../../../SpacesModule/SpacesService/doSpacesService";
import { RetailerServiceRepository } from "./retailer-service.repository";
import { CreateRetailerServiceDto } from "./dto/create-retailer-service.dto";
import { RetailerService } from "./entities/retailer-service.entity";
import { UploadedMulterFileI } from "../../../SpacesModule/SpacesService";
import { ElasticsearchService } from "@nestjs/elasticsearch";
import { SearchRetailerDto } from "./entities/search-retailer.dto";
import { getGraphQLClient } from "src/helper/graphql-client-helper";
import { fetchAllRetailers } from "src/helper/dutchie-plus-query";

@Injectable()
export class RetailerServiceService {
    constructor(
        @InjectRepository(RetailerServiceRepository)
        private readonly repository: RetailerServiceRepository,
        private readonly doSpaceService: DoSpacesService,
        private readonly esService: ElasticsearchService
    ) {
    }

    async create(
      dto: CreateRetailerServiceDto
    ): Promise<RetailerService | Error> {
        const data = await this.repository.createEntity(dto);
        if (!data) {
            throw new HttpException(
              Constants.CREATE_FAILED,
              HttpStatus.BAD_REQUEST
            );
        }
        return ResponseUtils.successResponseHandler(201, "Data created successfully", "data", data);
    }

    async findAll(): Promise<RetailerService[] | Error> {
        const data = await this.repository.findByFilterQuery({});
        if (!data) {
            throw new HttpException(
                Constants.NOT_FOUND,
                HttpStatus.NOT_FOUND
            );
        }
        return ResponseUtils.successResponseHandler(200, "Data fetched successfully", "data", data);
    }

    async findOne(id: string): Promise<RetailerService[] | Error> {
        const data = await this.repository.findByFilterQuery({ productId: id });
        const reviewCount = Object.keys(data).length != 0 ? Object.keys(data).length : 0;
        if (!data) {
            throw new HttpException(
              Constants.NOT_FOUND,
              HttpStatus.NOT_FOUND
            );
        }
        let ratingAvarage = 0;
        if (Object.keys(data).length > 0) {
            let count = 0;
            for (let i = 0; i < reviewCount; i++) {
                count += parseInt(data[i].rating);
            }
            ratingAvarage = Math.round(count / reviewCount);
        }

        const response = {
            data: data,
            reviewCount: reviewCount,
            ratingAvarage: ratingAvarage
        };
        return ResponseUtils.successResponseHandler(200, "Data fetched successfully", "data", response);
    }

    async remove(id: string): Promise<boolean | Error> {
        const data = await this.repository.deleteEntity(id);
        if (!data) {
            throw new HttpException(
              Constants.NOT_FOUND,
              HttpStatus.NOT_FOUND
            );
        }
        return ResponseUtils.buildDeletedData(data, 200, "Data deleted successfully");
    }

    async update(
      id: string,
      dto: any,
      files: {
          retailerImage?: UploadedMulterFileI
      }
    ): Promise<any | Error> {
        if (files && files.retailerImage) {
            const retailerImage: any = await this.doSpaceService.uploadFile(files.retailerImage[0], "DutchiePlusRetailerService/");
            dto.retailerImage = retailerImage;
        }
        await this.repository.update(id, {
            retailerImage: dto.retailerImage,
            retailerMobileNumber: dto.retailerMobileNumber
        });
        const data = await this.repository.findOne({ where: { id: id } });
        if (!data) {
            throw new HttpException(
              Constants.UPDATE_FAILED,
              HttpStatus.BAD_REQUEST
            );
        }
        return ResponseUtils.successResponseHandler(200, "Data Updated successfully", "data", data);
    }

    async updateAlpineIq(id: string, dto: any): Promise<any | Error> {

        const FedexCutoffTime: any = await this.repository.findOneEntity(id);
        if (!FedexCutoffTime) {
            throw new HttpException(
              Constants.NOT_FOUND,
              HttpStatus.NOT_FOUND
            );
        }

        const updateData = await this.repository.update(id, {
            alpineIqApiKey: dto.alpineIqApiKey,
            alpineIqUID: dto.alpineIqUID
        });
        const data = await this.repository.findOne(id);
        if (!data) {
            throw new HttpException(
              Constants.NOT_FOUND,
              HttpStatus.NOT_FOUND
            );
        }

        return ResponseUtils.successResponseHandler(
          200,
          "Data Updated Successful!",
          "data",
          data);
    }

    public async indexIsExists(index: string) {
        return await this.esService.indices.exists({ index });
    }

    public async deleteIndex(index: string) {
        const isExists = await this.indexIsExists(index);
        if (isExists) {
            await this.esService.indices.delete({ index });
        }
    }

    public async createIndex(index: string, body: any) {
        const isExists = await this.indexIsExists(index);
        if (!isExists) {
            const data = await this.esService.indices.create({ index, body });
            return data.acknowledged;
        }
        return false;
    }

    public async indexData(index: string, body: any) {
        return await this.esService.index({ index, body });
    }

    private async getRetailerData() {
        const graphQLClient = getGraphQLClient();
        const query = fetchAllRetailers();
        const data = await graphQLClient.request(query);
        return data.retailers;
    }

    public async indexBulkData(indexValue: string) {
        console.log(">>>Start Bulk Insert Retailer Service<<<", new Date());
        // const res: any = await this.findAll();
        const retailers = await this.getRetailerData();
        // console.log(retailers);

        await this.deleteIndex(indexValue);

        const body = {
            settings: {
                analysis: {
                    filter: {
                        autocomplete_filter: {
                            type: "edge_ngram",
                            min_gram: 1,
                            max_gram: 15
                        }
                    },
                    analyzer: {
                        autocomplete: {
                            type: "custom",
                            tokenizer: "standard",
                            filter: ["lowercase", "autocomplete_filter"]
                        }
                    }
                }
            },

            mappings: {
                properties: {
                    retailerId: {
                        type: "text",
                        fields: {
                            keyword: {
                                type: "keyword",
                                ignore_above: 256
                            }
                        }
                    },
                    retailerName: {
                        type: "text",
                        analyzer: "autocomplete",
                        search_analyzer: "standard",
                        fields: {
                            keyword: {
                                type: "keyword",
                                ignore_above: 512
                            }
                        }
                    },
                }
            }
        };
        
        await this.createIndex(indexValue, body);

        const operations = retailers.flatMap((doc) => [
            { index: { _index: indexValue } },
            {
                retailerId: doc.id,
                retailerName: doc.name
            }
        ]);
        await this.esService.bulk({ refresh: true, operations });
        console.log(">>>End Bulk Insert RetailerService<<<", new Date());

        return ResponseUtils.successResponseHandler(200, "Elastic DataBase Update Successfully.");
    }

    public async search(index: string, searchDto: SearchRetailerDto) {
        const { search, limit } = searchDto;

        const result = await this.esService.search<any>({
            index,
            body: {
                query: {
                    bool: {
                        must: [
                            {
                                match: {
                                    retailerName: {
                                        query: search,
                                        fuzziness: "AUTO",
                                        fuzzy_transpositions: false,
                                        prefix_length: 2
                                    }
                                }
                            }
                        ]
                    }
                },
                highlight: {
                    fields: {
                        retailerName: {
                            pre_tags: ["<b>"],
                            post_tags: ["</b>"]
                        }
                    }
                }
            },
            size: limit && limit > 0 ? limit : 50
        });
        const hits = result.hits.hits;
        const pageCount: any = result.hits.total;

        const retailers = hits.map((item: any) => {
            return {
                ...item._source,
                highlight: item?.highlight?.retailerName?.[0]
            };
        });
        const data = { retailers, total: (pageCount?.value ?? pageCount) };

        if (!data) {
            throw new HttpException(Constants.NOT_FOUND, HttpStatus.NOT_FOUND);
        }
        return ResponseUtils.successResponseHandler(
            200,
            "Retailers list fetched successfully",
            "data",
            data
        );
    }
}
